/*
	SubSpace transmission control over core protocol by Catid@pacbell.net
*/

#ifndef HOST_H
#define HOST_H

#include "datatypes.h"
#include "sockets.h"
#include "chunk.h"
#include "encrypt.h"
#include "settings.h"

#define PACKET_SILENCE_LIMIT     2000
#define MAX_RELIABLE_IN_TRANSIT     3
#define CHUNK_SIZE                496
#define CLUSTER_MODE            false

#define HANDLER( a ) friend void a(hostMessage * msg);

class reliableMessage
{
public:
	char * msg;								// Message with the reliable header
	Uint32 len;								// Message length
	Uint32 ACK_ID;							// ACK identifier
	Uint32 lastSend;						// Last time it was sent

	void setTime();							// Update lastSend

	reliableMessage(Uint32 nACK_ID,
					char * nmsg,
					Uint32 nlen);			// Fill fields
	~reliableMessage();						// Free memory
};


class clusterMessage
{
public:
	char * msg;								// Message without the reliable header
	Uint32 len;								// Message length

	clusterMessage(char * nmsg,
				   Uint32 nlen);			// Fill fields
	~clusterMessage();						// Free memory
};


class hostMessage
{
public:
	char * msg;								// Trimmed buffer
	Uint32 len;								// Message length
	class Host * src;						// Sender of this message

	BYTE getType(bool special);				// Return type byte

	hostMessage(void * msg,
				Uint32 len,
				Host * src);				// Fill fields
	~hostMessage();							// Free memory
};


class Host
{
	/* Socket */
	UDPSocket * socket;						// Socket wrapper
	INADDR remote;							// sockaddr wrapper
	void send(char * msg, Uint32 len);		// Simply encrypt and send to host

	/* Inbox */
	Uint32 remoteStep;						// Next ACK_ID expected
	_linkedlist <reliableMessage> received;	// Messages waiting for a lost packet
	void checkReceived(Uint32 ACK_ID,
					   char * msg,
					   Uint32 len);			// Process queued messages or queue another
	void sendACK(Uint32 ACK_ID);			// Send an acknowledgement

	/* Outbox */
	Uint32 localStep;						// Next ACK_ID to be used
	Uint32 localSent;						// Last ACK_ID to be sent
	_linkedlist <reliableMessage> sent;		// Messages waiting for ACKnowledgement
	void checkSent(Uint32 ACK_ID);			// Route ACK messages here
	void sendQueued();						// Resend the lost messages

	_linkedlist <reliableMessage> queued;	// Send-queue overflows go here
	void sendNext();						// Start sending a queued message
	bool queue(char * msg, Uint32 len);		// Add to the queued list

	/* General Commands */
	_jumptable <hostMessage> generalRouter;	// Server/Client protocol

		HANDLER( handleSpecialHeader );		// 00 Call a core protocol function

		HANDLER( handleServerListResponse );// 01 Download list

		HANDLER( handleServerListRequest );	// 01 Upload list

	/* Special Commands */
	_jumptable <hostMessage> specialRouter;	// Core protocol

		HANDLER( handleEncryptRequest );	// 01 Client key & protocol version
		HANDLER( handleEncryptResponse );	// 02 Session encryption key

		HANDLER( handleReliable );			// 03 Process sequentially
		HANDLER( handleACK );				// 04 Reliable message acknowledgement

		HANDLER( handleSyncRequest );		// 05 May include a ploss check
		HANDLER( handleSyncResponse );		// 06 Produces TimeDiff

		HANDLER( handleDisconnect );		// 07 Session termination

		HANDLER( handleChunkBody );			// 08 Waits for a ChunkTail
		HANDLER( handleChunkTail );			// 09 Processes accumulated buffer
		HANDLER( handleBigChunk );			// 0A Buffer contains up to 20 MB

		HANDLER( handleCluster );			// 0E Carpool into one packet

	/* State */
	Uint32 lastRecv;						// Time since last packet was recv'd
	bool clustering;						// Manage send() timing
	bool serverSide;						// Enable server-side or client-side extensions?

	bool connecting;						// Until we get an 00 02, send 00 01
	Uint32 lastConnect;
	bool syncing;							// Until we get an 00 06, send 00 05
	Uint32 lastSync;

	/* Transfer modes */
	chunkBuffer little;						// 00 08 / 00 09
	chunkBuffer large;						// 00 0A
	Uint32 ftLength;						// Used to prevent DoS attacks (0 == no transfer)
	_linkedlist <clusterMessage> clustered;	// Clustered messages queued for sending
	void sendClustered(_listnode <clusterMessage> * head);

	/* Message encryption */
	SS_ENCR encryption;						// 00 01 / 00 02

	/* Synchronization */
	Uint32 msgSent;							// Total distinct messages sent
	Uint32 msgRecv;							// Total distinct messages recv'd
	Sint32 timeDiff;						// Delta T between server and client - changes over time
	Uint32 syncPing;						// Average host response time to sync requests
	Uint32 stdDev;							// Ping-time standard deviation

	Uint32 pingTable[32];					// Workspace to find the standard deviation in ping-times
	Uint32 tableIndex;						// Current table entry
	void resetPingTable();					// Initialize ping table to default values
	void addPing(Uint32 time);				// Add new data to the table; generate new stdDev

	/* Host communication */
	friend class hostList;
	bool killMe;							// Tells hostList to clean up

public:
	Host(bool protoSide, INADDR * hostName, UDPSocket * nSocket);
	~Host();
	void resetSession();

	// Route packets to the correct Host object
	bool validateSource(INADDR * src);							// Did a message originate here?
	void doEvents();

	// Carpool messages
	void beginCluster();										// Prepare clustered information
	void post(char * msg, Uint32 len, bool reliable);			// Add to clustered list
	void endCluster();											// Decide what to do with clustered messages

	// Synchronization
	Uint32 getHostTime();
	Uint32 getTime();

	// Routing
	void gotPacket(char * msg, Uint32 len);
	void gotMessage(char * msg, Uint32 len);
	void gotMessage(hostMessage * msg);
	void gotSpecialMessage(hostMessage * msg);

	// Core in
	void gotEncryptRequest(Uint32 key, Uint16 prot);
	void gotEncryptResponse(Uint32 key);
	void gotEncryptResponse(Uint32 key, BYTE mudge);
	void gotReliable(Uint32 id, char * msg, Uint32 len);
	void gotACK(Uint32 id);
	void gotSyncRequest();
	void gotSyncRequest(Uint32 time);
	void gotSyncRequest(Uint32 time, Uint32 sent, Uint32 recv);
	void gotSyncResponse(Uint32 pongTime, Uint32 pingTime);
	void gotDisconnect();
	void gotChunkBody(char * msg, Uint32 len);
	void gotChunkTail(char * msg, Uint32 len);
	void gotBigChunk(Uint32 total, char * msg, Uint32 len);
	void gotCluster(char * msg, Uint32 len);

	// Core out
	void connect();
	void disconnect(bool notify);
	void syncClocks();

	// Game protocol
	bool sendingList;
};


class hostList
{
public:
	_linkedlist <Host> list;
	Uint16 maximumHosts;

	UDPSocket socket;

	hostList();										// Client version
	hostList(Uint16 max_hosts, Uint16 local_port);	// Server version
	~hostList();

	Host * getSource(INADDR * remote);

	void doEvents();
	void connectHost(INADDR * remote);
	void connectHost(Uint32 ip, Uint16 port);

	void massDisconnect();
};

#include "clientprot.h"
#include "serverprot.h"

#endif	// HOST_H
